### Copyright (C) 2009 Vincent Legoll <vincent.legoll@gmail.com>

### Redistribution and use in source and binary forms, with or without
### modification, are permitted provided that the following conditions
### are met:
### 
### 1. Redistributions of source code must retain the above copyright
###    notice, this list of conditions and the following disclaimer.
### 2. Redistributions in binary form must reproduce the above copyright
###    notice, this list of conditions and the following disclaimer in the
###    documentation and/or other materials provided with the distribution.

### THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
### IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
### OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
### IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
### INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
### NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
### DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
### THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
### (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
### THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import subprocess

from . import svn

NULL = open(os.devnull, "w+b")


def cmdout(cmd, text=None, **kwargs):
    stdin = NULL
    if text is not None:
        stdin = subprocess.PIPE
    new_kwargs = {
                  'stdin': stdin,
                  'stdout': subprocess.PIPE,
                  'stderr': NULL,
                  }
    new_kwargs.update(kwargs)
    p = subprocess.Popen(cmd, **new_kwargs)
    out = p.communicate(text)[0]
    status = p.wait()
    return out, status


class Vc(svn.Vc):

    CMD = "svk"
    NAME = "SVK"

    def check_repo_root(self, location):
        try:
            # `svk info` may be interactive. It can ask to create its repository, we
            # don't want that to happen, so provide the right answer with `text="n"`
            status = cmdout([self.CMD, "info"], cwd=location, stdout=NULL,
                                 stderr=NULL, text='n')[1]
            # SVK does evil things to the real stdin, even when having its input
            # stream redirected to a newly created pipe, restore sanity manually
            cmdout(['stty', 'sane'], stdout=NULL, stderr=NULL, stdin=None)
        except OSError:
            raise ValueError()
        if status != 0:
            raise ValueError()
        return location
